//==============================================================================
// Alphamasking tutorial
// 2002-2004 by Sascha Willems
// http://www.delphigl.de
//==============================================================================
// Updated 19.01.2004 :
//  - Renamed TForm1 to TGLForm
//  - Replaced OpenGL12.pas with dglOpenGL.pas (GL1.5)
//  - DC and RC now released on destroy
//  - Fixed Clampmode (Replaced GL_CLAMP with GL_CLAMP_TO_EDGE)
//  - Moved texturebindings out of glbegin..glend
//==============================================================================

unit alphamask1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, dglOpenGL, glBMP, AppEvnts, ExtCtrls;

type
  TGLForm = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    Timer1: TTimer;
    Timer2: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure FormResize(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Timer1Timer(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
  public
    FDC    : HDC;
    FRC    : HGLRC;
    Frames : Integer;
    yrot   : TGLFloat;
    xrot   : TGLFloat;
    OMpos  : TPoint;
    procedure GLInit;
    procedure LoadTextures;
    procedure DrawScene;
    procedure GenerateSkyBox(pWidth, pHeight, pLength : TGLFloat);
   end;
  TVector = record
     x,y,z : TGLFLoat;
    end;
  TTree = object
    Position : TVector;
    Rotation : TGLFloat;
    Size     : TGLFloat;
    procedure Draw;
   end;

const
 NumTex    = 2;
 NumTrees  = 128;
 PiOver180 : TGLFloat = 0.0174532925;

var
  Texture       : Array[0..NumTex-1] of TGLBmp;
  Tree          : Array[0..NumTrees] of TTree;
  SkyBoxTexture : Array[0..5] of TGLBmp;
  PlayerPos     : TVector;
  Room          : Cardinal;
  SBox          : Cardinal;
  Started       : Boolean;
  GLForm        : TGLForm;

implementation

{$R *.dfm}

// Creates a random forest with NumTrees Trees
procedure CreateForest;
var
 i : Integer;
begin
Randomize;
for i := 0 to NumTrees do
 with Tree[i] do
  begin
  with Position do
   begin
   X := Random(20)*1.5+(Random(5)/10)-(Random(5)/10);
   Z := Random(20)*1.5+(Random(5)/10)-(Random(5)/10);
   if X < 0.25 then
    X := 0.25;
   if Y < 0.25 then
    Y := 0.25;
   Y := 0;
  end;
  Size     := 0.1;
  Rotation := Random(360);
  end;
end;

procedure TTree.Draw;
begin
// Enable blending
glEnable(GL_BLEND);
// Set the required blendfunction
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
// Enable Alphatesting (i.e. Alphamasking)
glEnable(GL_ALPHA_TEST);
// Set Alphafunction so that only Alphavalues greater 0.1 get through
glAlphaFunc(GL_GREATER, 0.1);

glPushMatrix;

glTranslatef(Position.x, Position.y, Position.z);
glRotatef(Rotation, 0, 1, 0);

Texture[1].Bind;

// Draw two quads with the alphamasked texture
glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(-0.5*Size, 0,      0);
  glTexCoord2f(0, 1); glVertex3f(-0.5*Size, 1*Size, 0);
  glTexCoord2f(1, 1); glVertex3f( 0.5*Size, 1*Size, 0);
  glTexCoord2f(1, 0); glVertex3f( 0.5*Size, 0,      0);

  glTexCoord2f(0, 0); glVertex3f(0, 0,      -0.5*Size);
  glTexCoord2f(0, 1); glVertex3f(0, 1*Size, -0.5*Size);
  glTexCoord2f(1, 1); glVertex3f(0, 1*Size,  0.5*Size);
  glTexCoord2f(1, 0); glVertex3f(0, 0,       0.5*Size);
glEnd;

glPopMatrix;
glPushMatrix;

// Draw the Shadow
glTranslatef(Position.x, Position.y, Position.z);
glRotatef(90, 0, 0, 1);
glColor4f(0,0,0,0.5);
glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(0.01, 0,      -0.5*Size);
  glTexCoord2f(0, 1); glVertex3f(0.01, 1.5*Size, -0.5*Size);
  glTexCoord2f(1, 1); glVertex3f(0.01, 1.5*Size,  0.5*Size);
  glTexCoord2f(1, 0); glVertex3f(0.01, 0.,      0.5*Size);
glEnd;
glColor4f(1,1,1,1);

glPopMatrix;
glDisable(GL_ALPHA_TEST);
glDisable(GL_BLEND);
end;

procedure TGLForm.FormCreate(Sender: TObject);
begin
Started := False;
if MessageDlg('Start in Fullscreenmode?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
 begin
 WindowState := wsMaximized;
 BorderStyle := bsNone;
 Cursor := crNone;
 end
else
 BorderStyle := bsSizeable;
InitOpenGL;
// Get a handle from our Window
FDC := GetDC(Handle);
// Generate our rendering context with a color depth of 32Bpp
FRC := CreateRenderingContext(FDC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
ActivateRenderingContext(FDC, FRC);
Frames := 0;
CreateForest;
GLInit;
LoadTextures;
PlayerPos.x := 15;
PlayerPos.y := 0.15;
PlayerPos.z := 15;
Started := True;
end;

// =============================================================================
// Sets up all needed OpenGL properties
// =============================================================================
procedure TGLForm.GLInit;
begin
glEnable(GL_TEXTURE_2D);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LEQUAL);
glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
end;

// =============================================================================
// Generates a Displaylist for the skybox
// =============================================================================
procedure TGLForm.GenerateSkyBox(pWidth, pHeight, pLength : TGLFloat);
var
 px,py,pz : TGLFloat;
begin
SBox := glGenLists(1);
glNewList(SBox, GL_COMPILE);
 px := - pWidth  / 2;
 py := - pHeight / 2;
 pz := - pLength / 2;
 // Back
 SkyBoxTexture[0].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(px,          py,           pz);
  glTexCoord2f(0, 1); glVertex3f(px,          py + pHeight, pz);
  glTexCoord2f(1, 1); glVertex3f(px + pWidth, py + pHeight, pz);
  glTexCoord2f(1, 0); glVertex3f(px + pWidth, py,           pz);
 glEnd;
 // Front
 SkyBoxTexture[1].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(1, 0); glVertex3f(px,	      py,           pz + pLength);
  glTexCoord2f(1, 1); glVertex3f(px,          py + pHeight, pz + pLength);
  glTexCoord2f(0, 1); glVertex3f(px + pWidth, py + pHeight, pz + pLength);
  glTexCoord2f(0, 0); glVertex3f(px + pWidth, py,           pz + pLength);
 glEnd;
 // Bottom
 SkyBoxTexture[2].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex3f(px + pWidth,	      py, pz);
  glTexCoord2f(1, 0); glVertex3f(px + pWidth,	      py, pz + pLength);
  glTexCoord2f(0, 0); glVertex3f(px, py, pz + pLength);
  glTexCoord2f(0, 1); glVertex3f(px, py, pz);
 glEnd;
 // Top
 SkyBoxTexture[3].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(px,          py + pHeight, pz);
  glTexCoord2f(0, 1); glVertex3f(px,          py + pHeight, pz + pLength);
  glTexCoord2f(1, 1); glVertex3f(px + pWidth, py + pHeight, pz + pLength);
  glTexCoord2f(1, 0); glVertex3f(px + pWidth, py + pHeight, pz);
 glEnd;
 // Left
 SkyBoxTexture[4].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(1, 0); glVertex3f(px, py,           pz);
  glTexCoord2f(0, 0); glVertex3f(px, py,           pz + pLength);
  glTexCoord2f(0, 1); glVertex3f(px, py + pHeight, pz + pLength);
  glTexCoord2f(1, 1); glVertex3f(px, py + pHeight, pz);
 glEnd;
 // Right
 SkyBoxTexture[5].Bind;
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(px + pWidth, py,           pz);
  glTexCoord2f(1, 0); glVertex3f(px + pWidth, py,           pz + pLength);
  glTexCoord2f(1, 1); glVertex3f(px + pWidth, py + pHeight, pz + pLength);
  glTexCoord2f(0, 1); glVertex3f(px + pWidth, py + pHeight, pz);
 glEnd;
glEndList;
end;

// =============================================================================
// Loads all needed textures
// =============================================================================
procedure TGLForm.LoadTextures;
const
 Name       : array[0..NumTex-1] of String = ('grass1', 'tree');
 SkyBoxName : array[0..5] of String = ('BK', 'FR', 'DN', 'UP', 'LF', 'RT');
var
 i : Integer;
begin
for i := 0 to High(Name) do
 begin
 Texture[i] := TGLBmp.Create;
 Texture[i].LoadImage('tex\'+Name[i]+'.jpg');
 // if we've loaded the tree texture, then we also load it's alpha channel
 if Name[i] = 'tree' then
  if not Texture[i].AddAlpha('tex\treealpha.jpg') then
   ShowMessage('Couldn''t load Alphamask!');
 Texture[i].GenTexture(False,False);
 end;
// Now we load the images and generate our textures
for i := 0 to High(SkyBoxTexture) do
 begin
 SkyBoxTexture[i] := TGLBmp.Create;
 SkyBoxTexture[i].LoadImage(SkyBoxName[i]+'.jpg');
 // We need to clamp our textures so that our borders look right
 SkyBoxTexture[i].SetTextureWrap(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
 SkyBoxTexture[i].GenTexture(False, False);
 end;
GenerateSkyBox(64, 64, 64);
end;

// =============================================================================
// Draws the Scene
// =============================================================================
procedure TGLForm.DrawScene;
var
 i : Integer;
begin
glMatrixMode(GL_MODELVIEW);
glClear(GL_DEPTH_BUFFER_BIT);
glLoadIdentity;
glRotatef(xrot, 1 ,0, 0);
glRotatef(yrot, 0 ,1, 0);

// Call the display list of the skybox
glCallList(SBox);

glTranslatef(-PlayerPos.x, -PlayerPos.y, -PlayerPos.z);

// Draw the floor
Texture[0].Bind;
glBegin(GL_QUADS);
 glTexCoord2f( 0,  0); glVertex3f(  0, 0,   0);
 glTexCoord2f( 0, 10); glVertex3f(  0, 0,  30);
 glTexCoord2f(10, 10); glVertex3f( 30, 0,  30);
 glTexCoord2f(10,  0); glVertex3f( 30, 0,   0);
glEnd;

// Draw the trees
for i := 0 to NumTrees do
 Tree[i].Draw;

SwapBuffers(FDC);
end;

procedure TGLForm.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
begin
if not Started then
 exit;
DrawScene;
Done := False;
inc(Frames);
end;

procedure TGLForm.FormResize(Sender: TObject);
begin
if not Started then
 exit;
if HandleAllocated then
 begin
 glViewport(0, 0, ClientWidth, ClientHeight);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(45, ClientWidth/ClientHeight, 0.1, 1000);
 end;
end;

procedure TGLForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
var
 NMPos : TPoint;
begin
NMPos := Mouse.CursorPos;
if OMPos.X < NMPos.X then
 yrot := yrot+Abs(OMPos.X-NMPos.X)/2;
if OMPos.X > NMPos.X then
 yrot := yrot-Abs(OMPos.X-NMPos.X)/2;
if (OMPos.Y>NMPos.Y) and (xrot>-60) then
 xrot := xrot-Abs(OMPos.Y-NMPos.Y)/4;
if (OMPos.Y<NMPos.Y) and (xrot<60) then
 xrot := xrot+Abs(OMPos.Y-NMPos.Y)/4;
OMPos := NMPos;
if NMPos.X < Left+10 then SetCursorPos(Left+ClientWidth-11, NMPos.Y);
if NMPos.X >= Left+ClientWidth-10 then SetCursorPos(Left+11, NMPos.Y);
if NMPos.Y < Top+40 then SetCursorPos(NMPos.X, Top+ClientHeight-11);
if NMPos.Y >= Top+ClientHeight-10 then SetCursorPos(NMPos.X, Top+31);
NMPos := Mouse.CursorPos;
OMPos := NMPos;
end;

procedure TGLForm.Timer1Timer(Sender: TObject);
var
 i : Integer;
begin
// Randomly increase the size of the trees to create the effect of a growing forest
for i := 0 to NumTrees do
 if Random(3) = 1 then
  if Tree[i].Size < 2 then
   Tree[i].Size := Tree[i].Size + Random(10)/400;
end;

procedure TGLForm.Timer2Timer(Sender: TObject);
begin
Caption := 'www.DelphiGL.de -Tutorial no.4 - Alphamasking - '+IntToStr(Frames)+' FPS';
Frames := 0;
end;

procedure TGLForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
case Key of
 VK_UP   : begin
           PlayerPos.x := PlayerPos.x-sin(-yrot*piover180)*0.1;
           PlayerPos.z := PlayerPos.z-cos(-yrot*piover180)*0.1;
           end;
 VK_DOWN : begin
           PlayerPos.x := PlayerPos.x+sin(-yrot*piover180)*0.1;
           PlayerPos.z := PlayerPos.z+cos(-yrot*piover180)*0.1;
           end;
end;
end;

procedure TGLForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
if Key = #27 then
 Close;
end;

procedure TGLForm.FormDestroy(Sender: TObject);
begin
DeActivateRenderingContext;
wglDeleteContext(FRC);
ReleaseDC(Handle, FDC);
end;

end.
